"""SoCo-CLI is a command line control interface for Sonos systems.

It is a simplified wrapper around the SoCo python library, as well as providing
an extensive range of additional features.

It can be used as a command line program, as an interactive command shell, and
in other programs via its simple API. It can also run as a simple HTTP API
server, to control Sonos via HTTP requests.

For more information, please see: https://github.com/avantrec/soco-cli
"""

__version__ = "0.4.14"
